from __future__ import unicode_literals
from django.conf.urls import *

from .views import datatable_manager

app_name = 'django_datatables'

urlpatterns = [
    url(r'^data/$', datatable_manager, name="datatable_manager")
]



from datetime import datetime
import logging

from django.core.serializers.json import DjangoJSONEncoder
from django.http import JsonResponse
from django.utils.encoding import force_text
from django.utils.functional import Promise
from django.utils.translation import ugettext as _
from django.utils.cache import add_never_cache_headers

try:
    from .excel import ExcelWriter
except ImportError:
    ExcelWriter = None

LOG = logging.getLogger(__name__)


class LazyEncoder(DjangoJSONEncoder):
    """Encodes django's lazy i18n strings
    """

    def default(self, obj):
        if isinstance(obj, Promise):
            return force_text(obj)
        return super(LazyEncoder, self).default(obj)


class DataResponse(object):

    def create_excel_response(self, request):
        """
        Return an excel writer as a response.
        """
        headers = self.get_column_titles()
        rows = self.get_data(request)
        title = getattr(self._meta, "title", "Sheet")

        xlwriter = ExcelWriter()
        xlwriter.add_headers(title, headers)
        for row in rows:
            xlwriter.add_row(title, dict(zip(headers, row)))

        return xlwriter.download(f'{title}-{datetime.now().strftime("%Y-%m-%d %H%m")}.xlsx')

    def create_data_response(self, func_val, request):
        try:
            assert isinstance(func_val, dict)
            response = dict(func_val)
            if 'result' not in response:
                response['result'] = 'ok'
        except KeyboardInterrupt:
            # Allow keyboard interrupts through for debugging.
            raise
        except Exception as e:
            LOG.exception('JSON view error: %s', request.path)
            msg = getattr(e, 'message', _('Internal error') + ': ') + str(e)
            response = {'result': 'error', 'sError': msg, 'text': msg}

        return JsonResponse(response)

    def dispatch(self, request, *args, **kwargs):
        self.request = request
        response = None

        if request.GET.get("export") == "excel":
            return self.create_excel_response(request)

        func_val = self.get_context_data(request)
        response = self.create_data_response(func_val, request)

        add_never_cache_headers(response)
        return response



from collections import OrderedDict
from .column import *


class AttrDict(dict):
    """A dictionary with attribute-style access. It maps attribute access to
    the real dictionary.  """

    def __init__(self, init={}):
        dict.__init__(self, init)

    def __getstate__(self):
        return self.__dict__.items()

    def __setstate__(self, items):
        for key, val in items:
            self.__dict__[key] = val

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, dict.__repr__(self))

    def __setitem__(self, key, value):
        return super(AttrDict, self).__setitem__(key, value)

    def __getitem__(self, name):
        return super(AttrDict, self).__getitem__(name)

    def __delitem__(self, name):
        return super(AttrDict, self).__delitem__(name)

    __getattr__ = __getitem__
    __setattr__ = __setitem__


def collect_current_columns(attrs):
    """ Collect fields/columns from current class. """
    current_columns = []
    for key, value in list(attrs.items()):
        if isinstance(value, Column):
            current_columns.append((key, value))
            del attrs[key]

    current_columns.sort(key=lambda x: x[1].creation_counter)
    attrs['declared_fields'] = OrderedDict(current_columns)

    return attrs


def assign_meta(new_class, bases, meta):
    m = {}
    for base in bases:
        m.update({k: v for k, v in getattr(base, "_meta", {}).items()})

    m.update({k: v for k, v in getattr(meta, "__dict__", {}).items() if not k.startswith("__")})
    _meta = AttrDict(m)

    return _meta


def filter_declared_fields(base, declared_fields):
    for attr, value in base.__dict__.items():
        if value is None and attr in declared_fields:
            del declared_fields[attr]
    return declared_fields


def define_declared_fields(new_class):
    declared_fields = OrderedDict()

    for base in reversed(new_class.__mro__):

        # Collect fields from base class.
        if hasattr(base, 'declared_fields'):
            declared_fields.update(base.declared_fields)

        # Field shadowing.
        declared_fields = filter_declared_fields(base, declared_fields)

    return declared_fields


class DeclarativeFieldsMetaclass(type):
    """
    Metaclass that collects Fields declared on the base classes.
    """
    def __new__(mcs, name, bases, attrs):

        # Pop the Meta class if exists
        meta = attrs.pop('Meta', None)

        attrs = collect_current_columns(attrs)

        new_class = (super(DeclarativeFieldsMetaclass, mcs).__new__(mcs, name, bases, attrs))

        _meta = assign_meta(new_class, bases, meta)

        # Walk through the MRO.
        declared_fields = define_declared_fields(new_class)

        new_class.base_fields = declared_fields
        new_class.declared_fields = declared_fields
        new_class._meta = _meta

        return new_class



"""
Column classes
"""

from django.urls import reverse


class Column(object):

    # Tracks each time a Field instance is created. Used to retain order.
    creation_counter = 0

    def __init__(self, title=None, css_class=None, value=None, link=None, link_args=None):
        self.title = title
        self.value = value
        self.link = link
        self.css_class = css_class
        self.link_args = link_args or []

        # Increase the creation counter, and save our local copy.
        self.creation_counter = Column.creation_counter
        Column.creation_counter += 1

    def render_column(self, value):
        """
        Returns a rendered value for the field.
        """
        return value

    def render_column_using_values(self, value, values_dict):
        """
        Return a rendered value which need to reference the Model's values dict
        """
        return value

    def get_referenced_values(self):
        """ Returns a list of values that will need to be referenced """
        values = []
        if self.has_link():  # link will need link_args
            for link_arg in self.link_args:
                if link_arg[0] not in (".", "#"):
                    values.append(link_arg)

        if type(self) == CheckBoxColumn:  # CheckBoxColumn's "value" is a field
            values.append(self.value)
        return values

    def render_link(self, value, values_dict):
        """
        Returns value wrapped in link tag as specified by link
        """
        def get_link_val(key):
            """ Gets a link by column, fixed string (#), or class attribute (.)"""
            if key.startswith("#"):
                return key[1:]
            elif key.startswith("."):
                return getattr(self, key[1:])
            return values_dict[key]

        reverse_args = [get_link_val(key) for key in self.link_args]
        link = reverse(self.link, args=reverse_args)
        return '<a href="{link}">{val}</a>'.format(link=link, val=value)

    def has_link(self):
        """ Returns True if column has link property set """
        return self.link is not None


class TextColumn(Column):
    pass


class CheckBoxColumn(Column):

    def __init__(self, name=None, *args, **kwargs):
        self.name = name
        self.db_independant = True
        super(CheckBoxColumn, self).__init__(*args, **kwargs)

    def render_column_using_values(self, value, values_dict):
        return '<input id="{id}" type="checkbox" name="{name}" value="{value}"></>'.format(
            id='???',
            name=self.name if self.name else '',
            value=values_dict[self.value],
        )


class GlyphiconColumn(Column):

    def __init__(self, icon, *args, **kwargs):
        self.icon = icon
        self.db_independant = True
        super(GlyphiconColumn, self).__init__(*args, **kwargs)

    def render_column(self, value):
        return "<span class='glyphicon glyphicon-{}'></span>".format(self.icon)


class FontAwesome4Column(Column):

    def __init__(self, icon, *args, **kwargs):
        if icon.startswith('fa-'):
            icon = icon[3:]
        self.icon = icon
        self.db_independant = True
        super(FontAwesome4Column, self).__init__(*args, **kwargs)

    def render_column(self, value):
        return """<i class="fa fa-{}" aria-hidden="true"></i>""".format(self.icon)


class FontAwesome5Column(Column):

    def __init__(self, icon, *args, **kwargs):
        self.icon = icon
        self.db_independant = True
        super(FontAwesome5Column, self).__init__(*args, **kwargs)

    def render_column(self, value):
        return """<i class="{}"></i>""".format(self.icon)


class BulletedListColumn(Column):

    def render_column(self, value):
        items = '\n'.join(map(lambda v: f'<li>{v}</li>', value))

        return f"""
            <ul style="margin: 0; padding-left: 1.5em;">
                {items}
            </ul>
        """


class ConstantTextColumn(Column):

    def __init__(self, text, *args, **kwargs):
        self.text = text
        self.db_independant = True
        super(ConstantTextColumn, self).__init__(*args, **kwargs)

    def render_column(self, value):
        return self.text


class DateColumn(Column):
    """
    Renders a date in Y-m-d format
    """
    def render_column(self, value):
        if value:
            return value.strftime("%Y-%m-%d").upper()
        return ''


class StringColumn(Column):
    """
    Does not ask the database for a column.  Used to custom render
    values from other columns.
    """
    def __init__(self, *args, **kwargs):
        self.db_independant = True
        super(StringColumn, self).__init__(*args, **kwargs)



from .views import *

import importlib


def datatable_manager(request):
    """
    Return the json data for a datatable
    """
    module_name = request.GET.get("module")
    name = request.GET.get("name")
    if not module_name or not name:
        return None
    module = importlib.import_module(module_name)
    cls = getattr(module, name)
    instance = cls()
    instance.request = request
    view_method = instance.dispatch
    return view_method(request)



"""
Datatable classes
"""

from collections import OrderedDict
import logging
from json import dumps
import sys

from pyquerystring import parse

from django.conf import settings
from django.db.models import Q
from django.utils import six
from django.utils.safestring import mark_safe
from django.views.debug import ExceptionReporter
from django.template.loader import select_template

from .column import *
from .mixins import DataResponse
from .datatable_meta import DeclarativeFieldsMetaclass

LOG = logging.getLogger(__name__)


class DatatableBase(six.with_metaclass(DeclarativeFieldsMetaclass)):
    """ JSON data for datatables
    """
    class Meta:
        order_columns = []
        # max limit of records returned, do not allow to kill our server by huge sets of data
        max_display_length = 100
        extra_fields = []
        searching = False

    @property
    def _querydict(self):
        if self.request.method == 'POST':
            return self.request.POST

        return parse(self.request.GET)

    def get_column_titles(self):
        """ Return list of column titles for the template engine
        """
        titles = []
        for key, column in self.declared_fields.items():
            titles.append(column.title if column.title else key.replace("_", " ").title())
        return titles

    def get_values_list(self):
        """
        Returns a list of the values to retrieve from the ORM.
        Do not return columns marked as db_independant.
        """
        values = [
            c.value if c.value else k for k, c in self.declared_fields.items(
            ) if not getattr(c, 'db_independant', False)
        ]
        values.extend(self._meta.extra_fields)
        return values

    def get_referenced_values(self):
        """
        Returns a list of values to retrieve
        The values will need to be referenced but might not be displayed
        """
        referenced_values = []
        for column in self.declared_fields.values():
            referenced_values += column.get_referenced_values()
        return referenced_values

    def _render_column(self, field, column, ic):
        # Call the render_column method for each field
        for ir, row_dict in enumerate(self.values_dicts):
            self.rendered_columns[ir][ic] = column.render_column(row_dict.get(field))
            self.rendered_columns[ir][ic] = column.render_column_using_values(
                self.rendered_columns[ir][ic], row_dict)

    def execute_method(self, method, column, ic):
        for ir, row_dict in enumerate(self.values_dicts):
            if getattr(column, 'db_independant', False):
                # send row to db_indepenants
                self.rendered_columns[ir][ic] = method(row_dict)
            else:
                self.rendered_columns[ir][ic] = method(self.rendered_columns[ir][ic])

    def _render_field(self, field, column, ic):
        """ Call the render_{} method for each column in local class """
        method_name = "render_{}".format(field)
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            if callable(method):
                self.execute_method(method, column, ic)

    def _render_link(self, column, ic):
        """ Call the render_link to wrap column in link tag """
        for ir, row_dict in enumerate(self.values_dicts):
            rendered_columns[ir][ic] = column.render_link(
                rendered_columns[ir][ic], row_dict)

    def render_columns(self):
        """
        Renders a column on a row
        """
        fields = self.declared_fields.keys()
        self.rendered_columns = [[None] * len(fields) for i in range(len(self.values_dicts))]

        for ic, field in enumerate(fields):
            column = self.declared_fields[field]
            if column.value:
                field = column.value

            self._render_column(field, column, ic)
            self._render_field(field, column, ic)

            if column.has_link():
                self._render_link(column, ic)

        return self.rendered_columns

    def ordering(self, qs):
        """
        Get parameters from the request and prepare order by clause
        """
        order = []

        sorting_cols = self._querydict.get('order', {})
        for info in sorting_cols:
            column_index = int(info['column'])
            declared_fields_keys = list(self.declared_fields.keys())
            field = declared_fields_keys[column_index]
            column_key = self.declared_fields[field].value or field

            sort_dir = '-' if info['dir'] == 'desc' else ''
            order.append('{0}{1}'.format(sort_dir, column_key))

        if order:
            return qs.order_by(*order)
        return qs

    def paging(self, qs):
        """ Paging
        """
        limit = min(int(self._querydict.get('length', 25)), self._meta.max_display_length)
        start = int(self._querydict.get('start', 0))

        # if pagination is disabled ("paging": false)
        if limit == -1:
            return qs

        offset = start + limit

        return qs[start:offset]

    def get_initial_queryset(self, request):
        if not self._meta.model:
            raise NotImplementedError("Need to provide a model or implement get_initial_queryset!")
        return self._meta.model.objects.all()

    def filter_through_field_lookup(self, search):
        field_lookup_suffixes = ('exact', 'contains', 'startswith',
                                 'endswith', 'search', 'regex')

        for field_lookup in self._meta.search_fields:
            if not field_lookup.endswith(field_lookup_suffixes):
                    # if no suffix provided, append "__icontains"
                field_lookup += '__icontains'
            q |= Q(**{field_lookup: search})

        return q

    def filter_by_search(self, qs):
        """
        Filter queryset as specified by search_fields
        Searches on icontains unless otherwise specified
        """
        search = self.request.GET.get('search[value]', None)
        if not search:
            return qs

        if getattr(self._meta, 'search_min_length', 0) <= len(search) and hasattr(
                self._meta, "search_fields"):
            q = Q()
            q = self.filter_through_field_lookup(search)
            qs = qs.filter(q)

        return qs

    def prepare_results(self, qs):
        values_to_get = set(self.get_values_list())
        values_to_get = values_to_get.union(set(self.get_referenced_values()))
        self.values_dicts = qs.values(*values_to_get)

        rendered_values = self.render_columns()
        data = [row for row in rendered_values]
        return data

    def get_data(self, request):
        """
        Gets all data, unpaged, as a list of dicts.
        """
        try:
            qs = self.get_initial_queryset(request)
            qs = self.filter_by_search(qs)
            qs = self.ordering(qs)
            data = self.prepare_results(qs)
        except Exception as e:
            LOG.exception(str(e))
            data = {'error': self.report_traceback()}
        return data

    def get_context_data(self, request):
        """
        Gets paginated data.
        Returned as json dict.
        """

        json_response = dict(draw=0, recordsTotal=0, recordsFiltered=0, data=[])

        additional_data = request.GET.get("additional_data")
        filter_params = parse(additional_data) if additional_data else {}
        try:
            qs = self.get_initial_queryset(request)
            total_records = qs.count()
            qs = self.filter_by_search(qs)
            if filter_params:
                qs = qs.filter(**filter_params)

            # number of records after filtering
            total_display_records = qs.count()

            qs = self.ordering(qs)
            qs = self.paging(qs)
            data = self.prepare_results(qs)
            json_response.update({"draw": int(self._querydict.get('draw', 0)),
                                  "recordsTotal": total_records,
                                  "recordsFiltered": total_display_records,
                                  "data": data})

        except Exception as e:
            LOG.exception(str(e))
            json_response['error'] = self.report_traceback()

        return json_response

    def report_traceback(self):
        if settings.DEBUG:
            reporter = ExceptionReporter(None, *sys.exc_info())
            return "\n" + reporter.get_traceback_text()
        return "\nAn error occured while processing an AJAX request."


class Datatable(DatatableBase, DataResponse):

    def _config_columns(self):
        columns = []
        for key, column in self.declared_fields.items():
            column_config = {}
            if column.css_class:
                column_config['className'] = column.css_class
            if key not in self._meta.order_columns:
                column_config['orderable'] = False
            columns.append(column_config)

        return columns

    def _config_order(self):
        order = []
        for order_col in self._meta.initial_order:
            order_dir = 'asc'
            if order_col.startswith('-'):
                order_col = order_col[1:]
                order_dir = 'desc'
            declared_fields_keys = list(self.declared_fields.keys())
            order_index = declared_fields_keys.index(order_col)
            order.append([order_index, order_dir])

        return order

    def datatable_config(self):
        """
        Returns the json config for the datatables init method in javascript
        """
        config = {}

        config['columns'] = self._config_columns()
        config['searching'] = self._meta.searching
        config['order'] = self._config_order() if "initial_order" in self._meta else []

        # Initial Display Length
        config['iDisplayLength'] = self._meta.get('initial_rows_displayed', 25)
        config['serverSide'] = self._meta.get('server_side', True)

        return mark_safe(dumps(config))

    @property
    def filter_form(self):
        return self._meta.get('filter_form', None)

    def render(self):
        """
        Render the javascript and html to create the datatable
        """
        template = select_template(['django_datatables/table.html'])
        context = {
            "can_export_to_excel": self._meta.get('export_to_excel', False),
            "module": self.__module__,
            "name": self.__class__.__name__,
            "datatable": self,
        }
        template_content = template.render(context)
        return mark_safe(template_content)



import os
from pathlib import Path
import sys
from setuptools import setup

# allow setup.py to be run from any path
os.chdir(Path(__file__).absolute().parent)

if 'publish' in sys.argv:
    if 'test' in sys.argv:
        os.system('python setup.py sdist bdist_wheel upload -rtest')
    else:
        os.system('python setup.py sdist bdist_wheel upload')
    sys.exit()

setup()



import openpyxl
from openpyxl.styles import Style, Font, Color
from openpyxl.writer.dump_worksheet import ExcelDumpWriter

from collections import defaultdict, OrderedDict
from HTMLParser import HTMLParser
from io import BytesIO
from zipfile import ZipFile, ZIP_DEFLATED

from django.http import HttpResponse


class MLStripper(HTMLParser):
    def __init__(self):
        self.reset()
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return ''.join(self.fed)

def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()

class ExcelWriter(object):
    """
    Data structure of sheets with headers and data
    that allows for consistent export of data to excel.
    """

    def __init__(self):
        self.sheets = defaultdict(lambda: defaultdict(
            headers=list(),
            data=list(),
        ))
        self.workbook = openpyxl.Workbook(optimized_write=True)

    def add_headers(self, sheet_name, headers_list):
        """
        Add headers to receipt.
        :param sheet_name: name of sheet
        :param headers_list: list of column headers
        """
        self.sheets[sheet_name]["headers"].extend(headers_list)

    def set_data(self, all_data):
        """ Adds all data as {sheet_name:list(data)}"""
        for sheet_name, data in all_data.items():
            self.sheets[sheet_name]['data'] = data

    def set_sheet_data(self, sheet_name, data):
        """ Set a specific sheet to a list of data """
        self.sheets[sheet_name]['data'] = data

    def add_row(self, sheet_name, data_dict):
        """
        Append a row of data to a sheet
        :param sheet_name: name of sheet
        :param data_dict:
            dict where key, value --> col header name, row val to save
        """
        self.sheets[sheet_name]['data'].append(data_dict)

    def _get_row_style(self, row_dict):
        """ Override this function with logic
        that returns row style based on rec's contents
        :param row_dict: dict where key, value --> col header, row's value
        :return: None or an openpyxl.styles.Style
        """
        return None

    def _create_sheets(self):
        """
        Creates sheets in workbook using current sheets data
        """
        header_style = Style(Font(bold=True))

        for sheet_name, data in self.sheets.items():
            sheet = self.workbook.create_sheet(title=sheet_name)

            header_row = list()
            for header in data['headers']:
                header_row.append(dict(
                    value=header,
                    style=header_style,
                ))
            sheet.append(header_row)
            row_num = 1

            for rec in iter(data["data"]):
                if rec is None:
                    continue

                outrow = list()
                for index, key in enumerate(data['headers']):
                    try:
                        value = rec[key]
                    except KeyError:
                        value = ''
                    if isinstance(value, list):
                        value = '\r\n'.join(value)
                    if value is None:
                        value = ''
                    if not isinstance(value, unicode):
                        value = unicode(value).decode('utf-8')

                    value = strip_tags(value)

                    cell = dict(
                        value=value,
                        style=self._get_row_style(rec),
                    )

                    outrow.append(cell)

                sheet.append(outrow)
                row_num += 1

    def _save_virtual_workbook(self):
        """
        Return an in-memory workbook, suitable for a Django response.

        NOTE: Fixes openpyxl.writer.excel.save_virtual_workbook() for
        optimized workbooks. Package function uses ExcelWriter instead of
        ExcelDumpWriter with no way of specifying it.
        Had to copy the function and change which Writer it used.
        """
        writer = ExcelDumpWriter(self.workbook)
        temp_buffer = BytesIO()
        try:
            archive = ZipFile(temp_buffer, 'w', ZIP_DEFLATED)
            writer.write_data(archive)
        finally:
            archive.close()
        virtual_workbook = temp_buffer.getvalue()
        temp_buffer.close()
        return virtual_workbook

    def download(self, filename):
        """
        Return an HttpResponse downloading the workbook
        """
        self._create_sheets()

        response = HttpResponse(
            self._save_virtual_workbook(),
            content_type='application/vnd.ms-excel'
        )
        response['Content-Disposition'] = \
            'attachment; filename="{0}"'.format(filename)

        return response

    def save_to_file(self, filename):
        """"
        Save workbook to a file
        """
        self._create_sheets()
        self.workbook.save(filename)
